//! # Aws s3
//!
//! Aws s3 client for remotefs

use std::io::Read;
use std::path::{Path, PathBuf};
use std::sync::Arc;

use aws_config::Region;
use aws_config::default_provider::credentials::DefaultCredentialsChain;
use aws_config::meta::region::RegionProviderChain;
pub use aws_sdk_s3::Client as S3Client;
use aws_sdk_s3::config::{Builder as S3ClientBuilder, Credentials, ProvideCredentials};
use aws_sdk_s3::primitives::{ByteStream, SdkBody};
use aws_sdk_s3::types::{CompletedMultipartUpload, CompletedPart, Object};
#[cfg(target_os = "windows")]
use path_slash::PathExt as _;
use remotefs::fs::{Metadata, ReadStream, UnixPex, Welcome, WriteStream};
use remotefs::{File, RemoteError, RemoteErrorType, RemoteFs, RemoteResult};
use tokio::runtime::Runtime;

use super::object::S3Object;
use crate::utils::path as path_utils;

const MIN_MULTIPART_UPLOAD_SIZE: usize = 5 * 1024 * 1024; // 5MB

/// Aws s3 file system client
pub struct AwsS3Fs {
    client: Option<S3Client>,
    runtime: Arc<Runtime>,
    wrkdir: PathBuf,
    // -- options
    bucket_name: String,
    /// Region name, if unset `Custom`.
    region: Option<String>,
    /// Custom endpoint (useful for minio)
    endpoint: Option<String>,
    profile: Option<String>,
    access_key: Option<String>,
    secret_key: Option<String>,
    security_token: Option<String>,
    session_token: Option<String>,
    /// New path style. Required for some backends, such as MinIO
    new_path_style: bool,
}

#[derive(Debug)]
pub enum RemoteFsCredentials {
    Default(DefaultCredentialsChain),
    User(Credentials),
}

impl ProvideCredentials for RemoteFsCredentials {
    fn fallback_on_interrupt(&self) -> Option<Credentials> {
        match self {
            Self::Default(c) => c.fallback_on_interrupt(),
            Self::User(c) => c.fallback_on_interrupt(),
        }
    }

    fn provide_credentials<'a>(
        &'a self,
    ) -> aws_credential_types::provider::future::ProvideCredentials<'a>
    where
        Self: 'a,
    {
        match self {
            Self::Default(c) => c.provide_credentials(),
            Self::User(c) => c.provide_credentials(),
        }
    }
}

impl AwsS3Fs {
    /// Initialize a new `AwsS3Fs`
    pub fn new<S: AsRef<str>>(bucket: S, runtime: &Arc<Runtime>) -> Self {
        Self {
            client: None,
            runtime: runtime.clone(),
            wrkdir: PathBuf::from("/"),
            bucket_name: bucket.as_ref().to_string(),
            region: None,
            endpoint: None,
            profile: None,
            access_key: None,
            secret_key: None,
            security_token: None,
            session_token: None,
            new_path_style: false,
        }
    }

    /// Specify region to connect to
    pub fn region<S: AsRef<str>>(mut self, region: S) -> Self {
        self.region = Some(region.as_ref().to_string());
        self
    }

    /// Specify custom endpoint
    /// This should be used when trying to connect to `minio` or other API compatible endpoints.
    pub fn endpoint<S: AsRef<str>>(mut self, endpoint: S) -> Self {
        self.endpoint = Some(endpoint.as_ref().to_string());
        self
    }

    /// Set aws profile. If unset, "default" will be used
    pub fn profile<S: AsRef<str>>(mut self, profile: S) -> Self {
        self.profile = Some(profile.as_ref().to_string());
        self
    }

    /// Set whether to use new path style. Required for backends such as MinIO (Default: False)
    pub fn new_path_style(mut self, new_path_style: bool) -> Self {
        self.new_path_style = new_path_style;
        self
    }

    /// Specify access key for aws connection.
    /// If unset, will be read from environment variable `AWS_ACCESS_KEY_ID`
    pub fn access_key<S: AsRef<str>>(mut self, key: S) -> Self {
        self.access_key = Some(key.as_ref().to_string());
        self
    }

    /// Specify secret access key for aws connection.
    /// If unset, will be read from environment variable `AWS_SECRET_ACCESS_KEY`
    pub fn secret_access_key<S: AsRef<str>>(mut self, key: S) -> Self {
        self.secret_key = Some(key.as_ref().to_string());
        self
    }

    /// Specify security token for aws connection.
    /// If unset, will be read from environment variable `AWS_SECURITY_TOKEN`
    pub fn security_token<S: AsRef<str>>(mut self, key: S) -> Self {
        self.security_token = Some(key.as_ref().to_string());
        self
    }

    /// Specify session token for aws connection.
    /// If unset, will be read from environment variable `AWS_SESSION_TOKEN`
    pub fn session_token<S: AsRef<str>>(mut self, key: S) -> Self {
        self.session_token = Some(key.as_ref().to_string());
        self
    }

    // -- get ref

    /// Get a reference to the underlying aws sdk [`S3Client`] struct
    pub fn client(&self) -> Option<&S3Client> {
        self.client.as_ref()
    }

    // -- private

    /// List objects contained in `p` path
    fn list_objects(&self, p: &Path, list_dir: bool) -> RemoteResult<Vec<S3Object>> {
        // Make path relative
        let key: String = Self::fmt_path(p, list_dir);
        debug!("Query list directory {}; key: {}", p.display(), key);
        self.query_objects(key, true)
    }

    /// Stat an s3 object
    fn stat_object(&self, p: &Path) -> RemoteResult<S3Object> {
        let key: String = Self::fmt_path(p, false);
        debug!("Query stat object {}; key: {}", p.display(), key);
        let objects = self.query_objects(key, false)?;
        // Absolutize path
        let absol: PathBuf = path_utils::absolutize(Path::new("/"), p);
        // Find associated object
        match objects
            .into_iter()
            .find(|x| x.path.as_path() == absol.as_path())
        {
            Some(obj) => Ok(obj),
            None => Err(RemoteError::new_ex(
                RemoteErrorType::NoSuchFileOrDirectory,
                format!("{}: No such file or directory", p.display()),
            )),
        }
    }

    /// Query objects at key
    fn query_objects(
        &self,
        key: String,
        only_direct_children: bool,
    ) -> RemoteResult<Vec<S3Object>> {
        debug!("query objects with prefix: '{key}'");

        let fut = self
            .client
            .as_ref()
            .unwrap()
            .list_objects_v2()
            .bucket(self.bucket_name.as_str())
            .prefix(key.as_str())
            .send();
        let results = self.runtime.block_on(fut);
        match results {
            Ok(entries) => {
                let Some(contents) = entries.contents else {
                    debug!("No objects found at key {key}",);
                    return Ok(vec![]);
                };

                let objects: Vec<S3Object> = contents
                    .into_iter()
                    .filter(|object| {
                        if only_direct_children {
                            Self::list_object_should_be_kept(object, key.as_str())
                        } else {
                            true
                        }
                    })
                    .map(S3Object::from)
                    .collect();

                debug!("Found objects: {:?}", objects);
                Ok(objects)
            }
            Err(e) => Err(RemoteError::new_ex(RemoteErrorType::StatFailed, e)),
        }
    }

    /// Returns whether object should be kept after list command.
    /// The object won't be kept if:
    ///
    /// 1. is not a direct child of provided dir
    fn list_object_should_be_kept(obj: &Object, dir: &str) -> bool {
        Self::is_direct_child(obj.key.as_deref().unwrap_or_default(), dir)
    }

    /// Checks whether Object's key is direct child of `parent` path.
    fn is_direct_child(key: &str, parent: &str) -> bool {
        key == format!("{}{}", parent, S3Object::object_name(key))
            || key == format!("{}{}/", parent, S3Object::object_name(key))
    }

    /// Make s3 absolute path from a given path
    fn resolve(&self, p: &Path) -> PathBuf {
        path_utils::diff_paths(
            path_utils::absolutize(self.wrkdir.as_path(), p),
            Path::new("/"),
        )
        .unwrap_or_default()
    }

    /// fmt path for fsentry according to format expected by s3
    fn fmt_path(p: &Path, is_dir: bool) -> String {
        // prevent root as slash
        if p == Path::new("/") {
            return "".to_string();
        }
        // Remove root only if absolute
        #[cfg(target_family = "unix")]
        let is_absolute: bool = p.is_absolute();
        // NOTE: don't use is_absolute: on windows won't work
        #[cfg(target_family = "windows")]
        let is_absolute: bool = p.display().to_string().starts_with('/');
        let p: PathBuf = match is_absolute {
            true => path_utils::diff_paths(p, Path::new("/")).unwrap_or_default(),
            false => p.to_path_buf(),
        };
        // NOTE: windows only: resolve paths
        #[cfg(target_family = "windows")]
        let p: PathBuf = PathBuf::from(p.to_slash_lossy().to_string());
        // Fmt
        match is_dir {
            true => {
                let mut p: String = p.display().to_string();
                if !p.ends_with('/') {
                    p.push('/');
                }
                p
            }
            false => p.to_string_lossy().to_string(),
        }
    }

    /// Check connection status
    fn check_connection(&mut self) -> RemoteResult<()> {
        if self.is_connected() {
            Ok(())
        } else {
            Err(RemoteError::new(RemoteErrorType::NotConnected))
        }
    }

    /// Return whether connection should use anonymous credentials
    fn is_anonymous(&self) -> bool {
        self.access_key.is_none()
            && self.secret_key.is_none()
            && self.security_token.is_none()
            && self.session_token.is_none()
    }

    /// Load credentials for current session
    fn load_credentials(&self, region: RegionProviderChain) -> RemoteResult<RemoteFsCredentials> {
        if self.is_anonymous() {
            Ok(RemoteFsCredentials::Default(self.runtime.block_on(
                DefaultCredentialsChain::builder().region(region).build(),
            )))
        } else {
            let Some(access_key) = self.access_key.as_ref() else {
                return Err(RemoteError::new_ex(
                    RemoteErrorType::AuthenticationFailed,
                    "Access key not set",
                ));
            };
            let Some(secret_key) = self.secret_key.as_ref() else {
                return Err(RemoteError::new_ex(
                    RemoteErrorType::AuthenticationFailed,
                    "Secret key not set",
                ));
            };

            Ok(RemoteFsCredentials::User(Credentials::new(
                access_key,
                secret_key,
                self.session_token.clone(),
                None,
                "default",
            )))
        }
    }

    /// Initialize region to be used from connection parameters
    fn init_region(&self) -> RemoteResult<RegionProviderChain> {
        Ok(
            RegionProviderChain::first_try(self.region.as_ref().cloned().map(Region::new))
                .or_default_provider()
                .or_else(Region::new("us-west-2")),
        )
    }

    /// Make bucket based on current options
    fn make_client(
        &self,
        region: Option<Region>,
        credentials: impl ProvideCredentials + 'static,
    ) -> S3Client {
        let mut builder = S3ClientBuilder::new()
            .credentials_provider(credentials)
            .behavior_version_latest()
            .region(region);
        builder.set_force_path_style(Some(self.new_path_style));

        builder.set_endpoint_url(self.endpoint.clone());

        S3Client::from_conf(builder.build())
    }
}

impl RemoteFs for AwsS3Fs {
    fn connect(&mut self) -> RemoteResult<Welcome> {
        // Load credentials
        debug!("Loading credentials... (profile {:?})", self.profile);
        let region_provider = self.init_region()?;
        let region = self.runtime.block_on(region_provider.region());
        let credentials = self.load_credentials(region_provider)?;
        // Parse region
        trace!(
            "Parsing region: {}; endpoint: {}",
            self.region.as_deref().unwrap_or("NULL"),
            self.endpoint.as_deref().unwrap_or("NULL")
        );
        debug!(
            "Credentials loaded! Connecting to bucket {}...",
            self.bucket_name
        );
        self.client = Some(self.make_client(region, credentials));
        info!("Connection successfully established to s3 bucket");
        Ok(Welcome::default())
    }

    fn disconnect(&mut self) -> RemoteResult<()> {
        info!("Disconnecting from S3 bucket...");
        match self.client.take() {
            Some(bucket) => {
                drop(bucket);
                Ok(())
            }
            None => Err(RemoteError::new(RemoteErrorType::NotConnected)),
        }
    }

    fn is_connected(&mut self) -> bool {
        self.client.is_some()
    }

    fn pwd(&mut self) -> RemoteResult<PathBuf> {
        self.check_connection()?;
        Ok(self.wrkdir.clone())
    }

    fn change_dir(&mut self, dir: &Path) -> RemoteResult<PathBuf> {
        self.check_connection()?;
        // Always allow entering root
        if dir == Path::new("/") {
            self.wrkdir = dir.to_path_buf();
            debug!("New working directory: {}", self.wrkdir.display());
            return Ok(self.wrkdir.clone());
        }
        // Check if directory exists
        debug!("Entering directory {}...", dir.display());
        let dir_p: PathBuf = self.resolve(dir);
        let dir_s: String = Self::fmt_path(dir_p.as_path(), true);
        debug!("Searching for key {} (path: {})...", dir_s, dir_p.display());
        // Check if directory already exists
        if self
            .stat_object(PathBuf::from(dir_s.as_str()).as_path())
            .is_ok()
        {
            self.wrkdir = path_utils::absolutize(Path::new("/"), dir_p.as_path());
            debug!("New working directory: {}", self.wrkdir.display());
            Ok(self.wrkdir.clone())
        } else {
            Err(RemoteError::new(RemoteErrorType::NoSuchFileOrDirectory))
        }
    }

    fn list_dir(&mut self, path: &Path) -> RemoteResult<Vec<File>> {
        self.check_connection()?;
        self.list_objects(path, true)
            .map(|x| x.into_iter().map(|x| x.into()).collect())
    }

    fn stat(&mut self, path: &Path) -> RemoteResult<File> {
        self.check_connection()?;
        let path = self.resolve(path);
        if let Ok(obj) = self.stat_object(path.as_path()) {
            return Ok(obj.into());
        }
        // Try as a "directory"
        trace!("Failed to stat object as file; trying as a directory...");
        let path = PathBuf::from(Self::fmt_path(path.as_path(), true));
        self.stat_object(path.as_path()).map(|x| x.into())
    }

    fn setstat(&mut self, _path: &Path, _metadata: Metadata) -> RemoteResult<()> {
        Err(RemoteError::new(RemoteErrorType::UnsupportedFeature))
    }

    fn exists(&mut self, path: &Path) -> RemoteResult<bool> {
        match self.stat(path) {
            Ok(_) => Ok(true),
            Err(RemoteError {
                kind: RemoteErrorType::NoSuchFileOrDirectory,
                ..
            }) => Ok(false),
            Err(err) => Err(err),
        }
    }

    fn remove_file(&mut self, path: &Path) -> RemoteResult<()> {
        self.check_connection()?;
        let path = Self::fmt_path(self.resolve(path).as_path(), false);
        debug!("Removing object '{}'", path);
        let fut = self
            .client
            .as_ref()
            .unwrap()
            .delete_object()
            .bucket(self.bucket_name.as_str())
            .key(path.as_str())
            .send();

        self.runtime
            .block_on(fut)
            .map_err(|e| {
                RemoteError::new_ex(
                    RemoteErrorType::ProtocolError,
                    format!("Could not remove file: {}", e),
                )
            })
            .map(|_| ())
    }

    fn remove_dir(&mut self, path: &Path) -> RemoteResult<()> {
        self.check_connection()?;
        if !self.exists(path).ok().unwrap_or(false) {
            return Err(RemoteError::new(RemoteErrorType::NoSuchFileOrDirectory));
        }
        println!("{}", self.resolve(path).as_path().display());
        let path = Self::fmt_path(self.resolve(path).as_path(), true);
        debug!("Removing object {}...", path);
        let fut = self
            .client
            .as_ref()
            .unwrap()
            .delete_object()
            .bucket(self.bucket_name.as_str())
            .key(path.as_str())
            .send();

        self.runtime.block_on(fut).map(|_| ()).map_err(|e| {
            RemoteError::new_ex(
                RemoteErrorType::ProtocolError,
                format!("Could not remove directory: {}", e),
            )
        })
    }

    fn remove_dir_all(&mut self, path: &Path) -> RemoteResult<()> {
        debug!("Removing all content of {}", path.display());
        if self.remove_dir(path).is_err() {
            self.remove_file(path)
        } else {
            Ok(())
        }
    }

    fn create_dir(&mut self, path: &Path, _mode: UnixPex) -> RemoteResult<()> {
        self.check_connection()?;
        let dir: String = Self::fmt_path(self.resolve(path).as_path(), true);
        debug!("Making directory {}...", dir);
        // Check if directory already exists
        if self
            .stat_object(PathBuf::from(dir.as_str()).as_path())
            .is_ok()
        {
            error!("Directory {} already exists", dir);
            return Err(RemoteError::new(RemoteErrorType::DirectoryAlreadyExists));
        }
        let fut = self
            .client
            .as_ref()
            .unwrap()
            .put_object()
            .bucket(self.bucket_name.as_str())
            .key(dir.as_str())
            .send();

        self.runtime.block_on(fut).map(|_| ()).map_err(|e| {
            RemoteError::new_ex(
                RemoteErrorType::FileCreateDenied,
                format!("Could not make directory: {}", e),
            )
        })
    }

    fn symlink(&mut self, _path: &Path, _target: &Path) -> RemoteResult<()> {
        Err(RemoteError::new(RemoteErrorType::UnsupportedFeature))
    }

    fn copy(&mut self, _src: &Path, _dest: &Path) -> RemoteResult<()> {
        Err(RemoteError::new(RemoteErrorType::UnsupportedFeature))
    }

    fn mov(&mut self, _src: &Path, _dest: &Path) -> RemoteResult<()> {
        Err(RemoteError::new(RemoteErrorType::UnsupportedFeature))
    }

    fn exec(&mut self, _cmd: &str) -> RemoteResult<(u32, String)> {
        Err(RemoteError::new(RemoteErrorType::UnsupportedFeature))
    }

    fn append(&mut self, _path: &Path, _metadata: &Metadata) -> RemoteResult<WriteStream> {
        Err(RemoteError::new(RemoteErrorType::UnsupportedFeature))
    }

    fn create(&mut self, _path: &Path, _metadata: &Metadata) -> RemoteResult<WriteStream> {
        Err(RemoteError::new(RemoteErrorType::UnsupportedFeature))
    }

    fn open(&mut self, _path: &Path) -> RemoteResult<ReadStream> {
        Err(RemoteError::new(RemoteErrorType::UnsupportedFeature))
    }

    fn create_file(
        &mut self,
        path: &Path,
        _metadata: &Metadata,
        mut reader: Box<dyn Read + Send>,
    ) -> RemoteResult<u64> {
        self.check_connection()?;
        let src = self.resolve(path);
        let key = Self::fmt_path(src.as_path(), false);

        // we write 4096 bytes at a time
        debug!("Query PUT for key '{}'", key);
        let mut buf = vec![0; MIN_MULTIPART_UPLOAD_SIZE];
        let mut offset = 0;
        let mut part_number = 1;

        // create multipart upload
        let fut = self
            .client
            .as_ref()
            .unwrap()
            .create_multipart_upload()
            .bucket(self.bucket_name.as_str())
            .key(key.as_str())
            .send();

        let upload = self.runtime.block_on(fut).map_err(|e| {
            error!("Could not init multipart upload: {e:?}",);
            RemoteError::new_ex(
                RemoteErrorType::ProtocolError,
                format!("Could not init multipart upload: {}", e),
            )
        })?;
        let upload_id = upload.upload_id().ok_or(RemoteError::new_ex(
            RemoteErrorType::ProtocolError,
            "no multipart id",
        ))?;
        debug!("starting multipart upload with upload id {upload_id}");

        let mut upload_parts: Vec<aws_sdk_s3::types::CompletedPart> = Vec::new();

        loop {
            let buflen = reader.read(&mut buf).map_err(|e| {
                RemoteError::new_ex(
                    RemoteErrorType::IoError,
                    format!("Could not read file: {}", e),
                )
            })?;
            if buflen == 0 {
                break;
            }

            let bytestream = ByteStream::new(SdkBody::from(&buf[..buflen]));
            debug!("sending part {part_number} for {key} at offset {offset} for {buflen} bytes");

            let buflen = buflen as i64;

            let fut = self
                .client
                .as_ref()
                .unwrap()
                .upload_part()
                .bucket(self.bucket_name.as_str())
                .key(key.as_str())
                .upload_id(upload_id)
                .body(bytestream)
                .part_number(part_number)
                .send();

            let upload_part_res = self.runtime.block_on(fut).map_err(|e| {
                error!("Could not put file: {e:?}",);
                RemoteError::new_ex(
                    RemoteErrorType::ProtocolError,
                    format!("Could not put file: {}", e),
                )
            })?;

            upload_parts.push(
                CompletedPart::builder()
                    .e_tag(upload_part_res.e_tag.unwrap_or_default())
                    .part_number(part_number)
                    .build(),
            );

            debug!("written {buflen} bytes at offset {offset} for {key}");

            offset += buflen;
            part_number += 1;
        }

        // complete multipart
        let completed_multipart_upload: CompletedMultipartUpload =
            CompletedMultipartUpload::builder()
                .set_parts(Some(upload_parts))
                .build();

        let fut = self
            .client
            .as_ref()
            .unwrap()
            .complete_multipart_upload()
            .bucket(self.bucket_name.as_str())
            .key(key.as_str())
            .upload_id(upload_id)
            .multipart_upload(completed_multipart_upload)
            .send();

        self.runtime.block_on(fut).map_err(|e| {
            error!("Could not complete multipart upload: {e:?}",);
            RemoteError::new_ex(
                RemoteErrorType::ProtocolError,
                format!("Could not complete multipart upload: {}", e),
            )
        })?;

        Ok(offset as u64)
    }

    fn open_file(
        &mut self,
        src: &Path,
        mut dest: Box<dyn std::io::Write + Send>,
    ) -> RemoteResult<u64> {
        self.check_connection()?;
        if !self.exists(src).ok().unwrap_or(false) {
            return Err(RemoteError::new(RemoteErrorType::NoSuchFileOrDirectory));
        }
        let src = self.resolve(src);
        let key = Self::fmt_path(src.as_path(), false);
        info!("Query GET for key '{}'", key);

        let fut = self
            .client
            .as_ref()
            .unwrap()
            .get_object()
            .bucket(self.bucket_name.as_str())
            .key(key)
            .send();

        let mut data = self
            .runtime
            .block_on(fut)
            .map_err(|e| {
                RemoteError::new_ex(
                    RemoteErrorType::ProtocolError,
                    format!("Could not get file: {}", e),
                )
            })?
            .body;

        let mut n = 0;

        while let Some(bytes) = self.runtime.block_on(data.try_next()).map_err(|e| {
            RemoteError::new_ex(
                RemoteErrorType::ProtocolError,
                format!("Could not get file: {}", e),
            )
        })? {
            dest.write_all(&bytes).map_err(|e| {
                RemoteError::new_ex(
                    RemoteErrorType::IoError,
                    format!("Could not write file: {}", e),
                )
            })?;
            n += bytes.len() as u64;
        }

        Ok(n)
    }
}

#[cfg(test)]
mod test {

    #[cfg(feature = "with-s3-ci")]
    use std::env;
    #[cfg(any(feature = "with-s3-ci", feature = "with-containers"))]
    use std::io::Cursor;
    #[cfg(any(feature = "with-s3-ci", feature = "with-containers"))]
    use std::time::SystemTime;

    use pretty_assertions::assert_eq;

    use super::*;
    use crate::mock::container::Minio;

    #[test]
    fn should_init_s3() {
        let s3 = AwsS3Fs::new("aws-s3-test", &Arc::new(Runtime::new().unwrap()));
        assert_eq!(s3.wrkdir.as_path(), Path::new("/"));
        assert_eq!(s3.bucket_name.as_str(), "aws-s3-test");
        assert!(s3.region.is_none());
        assert!(s3.endpoint.is_none());
        assert_eq!(s3.is_anonymous(), true);
        assert_eq!(s3.new_path_style, false);
        assert!(s3.client.is_none());
        assert!(s3.access_key.is_none());
        assert!(s3.profile.is_none());
        assert!(s3.secret_key.is_none());
        assert!(s3.security_token.is_none());
        assert!(s3.session_token.is_none());
        assert!(s3.secret_key.is_none());
        assert!(s3.client.is_none());
    }

    #[test]
    fn should_init_s3_with_options() {
        let s3 = AwsS3Fs::new("aws-s3-test", &Arc::new(Runtime::new().unwrap()))
            .region("eu-central-1")
            .access_key("AKIA0000")
            .profile("default")
            .secret_access_key("PASSWORD")
            .security_token("secret")
            .session_token("token")
            .new_path_style(true)
            .endpoint("omar");
        assert_eq!(s3.bucket_name.as_str(), "aws-s3-test");
        assert_eq!(s3.region.as_deref().unwrap(), "eu-central-1");
        assert_eq!(s3.access_key.as_deref().unwrap(), "AKIA0000");
        assert_eq!(s3.secret_key.as_deref().unwrap(), "PASSWORD");
        assert_eq!(s3.security_token.as_deref().unwrap(), "secret");
        assert_eq!(s3.session_token.as_deref().unwrap(), "token");
        assert_eq!(s3.endpoint.as_deref().unwrap(), "omar");
        assert_eq!(s3.is_anonymous(), false);
        assert_eq!(s3.new_path_style, true);
    }

    #[test]
    fn s3_is_direct_child() {
        assert_eq!(AwsS3Fs::is_direct_child("pippo/", ""), true);
        assert_eq!(AwsS3Fs::is_direct_child("pippo/sottocartella/", ""), false);
        assert_eq!(
            AwsS3Fs::is_direct_child("pippo/sottocartella/", "pippo/"),
            true
        );
        assert_eq!(
            AwsS3Fs::is_direct_child("pippo/sottocartella/", "pippo"), // This case must be handled indeed
            false
        );
        assert_eq!(
            AwsS3Fs::is_direct_child("pippo/sottocartella/readme.md", "pippo/sottocartella/"),
            true
        );
        assert_eq!(
            AwsS3Fs::is_direct_child("pippo/sottocartella/readme.md", "pippo/sottocartella/"),
            true
        );
    }

    #[test]
    fn s3_resolve() {
        let mut s3 = AwsS3Fs::new("aws-s3-test", &Arc::new(Runtime::new().unwrap()));
        s3.wrkdir = PathBuf::from("/tmp");
        // Absolute
        assert_eq!(
            s3.resolve(Path::new("/tmp/sottocartella/")).as_path(),
            Path::new("tmp/sottocartella")
        );
        // Relative
        assert_eq!(
            s3.resolve(Path::new("subfolder/")).as_path(),
            Path::new("tmp/subfolder")
        );
    }

    #[test]
    fn s3_fmt_path() {
        assert_eq!(
            AwsS3Fs::fmt_path(Path::new("/tmp/omar.txt"), false).as_str(),
            "tmp/omar.txt"
        );
        assert_eq!(
            AwsS3Fs::fmt_path(Path::new("omar.txt"), false).as_str(),
            "omar.txt"
        );
        assert_eq!(
            AwsS3Fs::fmt_path(Path::new("/tmp/subfolder"), true).as_str(),
            "tmp/subfolder/"
        );
        assert_eq!(
            AwsS3Fs::fmt_path(Path::new("tmp/subfolder"), true).as_str(),
            "tmp/subfolder/"
        );
        assert_eq!(AwsS3Fs::fmt_path(Path::new("tmp"), true).as_str(), "tmp/");
        assert_eq!(AwsS3Fs::fmt_path(Path::new("tmp/"), true).as_str(), "tmp/");
        assert_eq!(AwsS3Fs::fmt_path(Path::new("/"), true).as_str(), "");
    }

    #[test]
    #[cfg(any(feature = "with-s3-ci", feature = "with-containers"))]
    fn should_not_append_to_file() {
        crate::mock::logger();
        let Ctx {
            mut client,
            container: _container,
        } = setup_client();
        // Create file
        let p = Path::new("a.txt");
        // Append to file
        let file_data = "Hello, world!\n";
        let reader = Cursor::new(file_data.as_bytes());
        assert!(
            client
                .append_file(p, &Metadata::default(), Box::new(reader))
                .is_err()
        );
        finalize_client(client);
    }

    #[test]
    #[cfg(any(feature = "with-s3-ci", feature = "with-containers"))]
    fn should_change_directory() {
        crate::mock::logger();
        let Ctx {
            mut client,
            container: _container,
        } = setup_client();
        let pwd = client.pwd().ok().unwrap();
        assert!(client.change_dir(Path::new("/")).is_ok());
        assert!(client.change_dir(pwd.as_path()).is_ok());
        finalize_client(client);
    }

    #[test]
    #[cfg(any(feature = "with-s3-ci", feature = "with-containers"))]
    fn should_not_change_directory() {
        crate::mock::logger();
        let Ctx {
            mut client,
            container: _container,
        } = setup_client();
        assert!(
            client
                .change_dir(Path::new("/tmp/sdfghjuireghiuergh/useghiyuwegh"))
                .is_err()
        );
        finalize_client(client);
    }

    #[test]
    #[cfg(any(feature = "with-s3-ci", feature = "with-containers"))]
    fn should_not_copy_file() {
        crate::mock::logger();
        let Ctx {
            mut client,
            container: _container,
        } = setup_client();
        // Create file
        let p = Path::new("a.txt");
        let file_data = "test data\n";
        let reader = Cursor::new(file_data.as_bytes());
        let mut metadata = Metadata::default();
        metadata.size = file_data.len() as u64;
        assert!(client.create_file(p, &metadata, Box::new(reader)).is_ok());
        assert!(client.copy(p, Path::new("aaa/bbbb/ccc/b.txt")).is_err());
        finalize_client(client);
    }

    #[test]
    #[cfg(any(feature = "with-s3-ci", feature = "with-containers"))]
    fn should_create_directory() {
        crate::mock::logger();
        let Ctx {
            mut client,
            container: _container,
        } = setup_client();
        // create directory
        assert!(
            client
                .create_dir(Path::new("mydir"), UnixPex::from(0o755))
                .is_ok()
        );
        finalize_client(client);
    }

    #[test]
    #[cfg(any(feature = "with-s3-ci", feature = "with-containers"))]
    fn should_not_create_directory_cause_already_exists() {
        crate::mock::logger();
        let Ctx {
            mut client,
            container: _container,
        } = setup_client();
        // create directory
        assert!(
            client
                .create_dir(Path::new("mydir"), UnixPex::from(0o755))
                .is_ok()
        );
        assert_eq!(
            client
                .create_dir(Path::new("mydir"), UnixPex::from(0o755))
                .err()
                .unwrap()
                .kind,
            RemoteErrorType::DirectoryAlreadyExists
        );
        finalize_client(client);
    }

    #[test]
    #[cfg(feature = "with-s3-ci")]
    fn should_not_create_directory() {
        crate::mock::logger();
        let Ctx {
            mut client,
            container: _container,
        } = setup_client();
        // create directory
        assert!(
            client
                .create_dir(
                    Path::new("/tmp/werfgjwerughjwurih/iwerjghiwgui"),
                    UnixPex::from(0o755)
                )
                .is_err()
        );
        finalize_client(client);
    }

    #[test]
    #[cfg(any(feature = "with-s3-ci", feature = "with-containers"))]
    fn should_create_file() {
        crate::mock::logger();
        let Ctx {
            mut client,
            container: _container,
        } = setup_client();
        // Create file
        let p = Path::new("a.txt");
        let file_data = "test data\n";
        let reader = Cursor::new(file_data.as_bytes());
        let mut metadata = Metadata::default();
        metadata.size = file_data.len() as u64;
        assert_eq!(
            client
                .create_file(p, &metadata, Box::new(reader))
                .ok()
                .unwrap(),
            10
        );
        // Verify size
        assert_eq!(client.stat(p).ok().unwrap().metadata().size, 10);
        finalize_client(client);
    }

    #[test]
    #[cfg(any(feature = "with-s3-ci", feature = "with-containers"))]
    fn should_not_exec_command() {
        crate::mock::logger();
        let Ctx {
            mut client,
            container: _container,
        } = setup_client();
        assert!(client.exec("echo 5").is_err());
        finalize_client(client);
    }

    #[test]
    #[cfg(any(feature = "with-s3-ci", feature = "with-containers"))]
    fn should_tell_whether_file_exists() {
        crate::mock::logger();
        let Ctx {
            mut client,
            container: _container,
        } = setup_client();
        // Create file
        let p = Path::new("a.txt");
        let file_data = "test data\n";
        let reader = Cursor::new(file_data.as_bytes());
        let mut metadata = Metadata::default();
        metadata.size = file_data.len() as u64;
        assert!(client.create_file(p, &metadata, Box::new(reader)).is_ok());
        // Verify size
        assert_eq!(client.exists(p).ok().unwrap(), true);
        assert_eq!(client.exists(Path::new("b.txt")).ok().unwrap(), false);
        assert_eq!(
            client.exists(Path::new("/tmp/ppppp/bhhrhu")).ok().unwrap(),
            false
        );
        finalize_client(client);
    }

    #[test]
    #[cfg(any(feature = "with-s3-ci", feature = "with-containers"))]
    fn should_list_dir() {
        crate::mock::logger();
        let Ctx {
            mut client,
            container: _container,
        } = setup_client();
        // Create file
        let wrkdir = client.pwd().ok().unwrap();
        let p = Path::new("a.txt");
        let file_data = "test data\n";
        let reader = Cursor::new(file_data.as_bytes());
        let mut metadata = Metadata::default();
        metadata.size = file_data.len() as u64;
        assert!(client.create_file(p, &metadata, Box::new(reader)).is_ok());
        // Verify size
        let file = client
            .list_dir(wrkdir.as_path())
            .ok()
            .unwrap()
            .get(0)
            .unwrap()
            .clone();
        assert_eq!(file.name().as_str(), "a.txt");
        let mut expected_path = wrkdir;
        expected_path.push(p);
        assert_eq!(file.path.as_path(), expected_path.as_path());
        assert_eq!(file.extension().as_deref().unwrap(), "txt");
        assert_eq!(file.metadata.size, 10);
        assert_eq!(file.metadata.mode, None);
        finalize_client(client);
    }

    #[test]
    #[cfg(any(feature = "with-s3-ci", feature = "with-containers"))]
    fn should_not_move_file() {
        crate::mock::logger();
        let Ctx {
            mut client,
            container: _container,
        } = setup_client();
        // Create file
        let p = Path::new("a.txt");
        let file_data = "test data\n";
        let reader = Cursor::new(file_data.as_bytes());
        let mut metadata = Metadata::default();
        metadata.size = file_data.len() as u64;
        assert!(client.create_file(p, &metadata, Box::new(reader)).is_ok());
        let dest = Path::new("b.txt");
        assert!(client.mov(p, dest).is_err());
        finalize_client(client);
    }

    #[test]
    #[cfg(any(feature = "with-s3-ci", feature = "with-containers"))]
    fn should_open_file() {
        crate::mock::logger();
        let Ctx {
            mut client,
            container: _container,
        } = setup_client();
        // Create file
        let p = Path::new("a.txt");
        let file_data = "test data\n";
        let reader = Cursor::new(file_data.as_bytes());
        let mut metadata = Metadata::default();
        metadata.size = file_data.len() as u64;
        assert!(client.create_file(p, &metadata, Box::new(reader)).is_ok());
        // Verify size
        let buffer: Box<dyn std::io::Write + Send> = Box::new(Vec::with_capacity(512));
        assert_eq!(client.open_file(p, buffer).ok().unwrap(), 10);
        finalize_client(client);
    }

    #[test]
    #[cfg(any(feature = "with-s3-ci", feature = "with-containers"))]
    fn should_write_big_data() {
        crate::mock::logger();
        let Ctx {
            mut client,
            container: _container,
        } = setup_client();
        // Create file
        let p = Path::new("a.txt");
        let file_data = vec![1; MIN_MULTIPART_UPLOAD_SIZE * 2];
        let reader = Cursor::new(file_data);
        let mut metadata = Metadata::default();
        metadata.size = (MIN_MULTIPART_UPLOAD_SIZE * 2) as u64;
        assert!(client.create_file(p, &metadata, Box::new(reader)).is_ok());
        // Verify size
        let buffer: Box<dyn std::io::Write + Send> = Box::new(Vec::with_capacity(512));
        assert_eq!(
            client.open_file(p, buffer).ok().unwrap(),
            (MIN_MULTIPART_UPLOAD_SIZE * 2) as u64
        );
        finalize_client(client);
    }

    #[test]
    #[cfg(any(feature = "with-s3-ci", feature = "with-containers"))]
    fn should_not_open_file() {
        crate::mock::logger();
        let Ctx {
            mut client,
            container: _container,
        } = setup_client();
        // Verify size
        let buffer: Box<dyn std::io::Write + Send> = Box::new(Vec::with_capacity(512));
        assert!(
            client
                .open_file(Path::new("/tmp/aashafb/hhh"), buffer)
                .is_err()
        );
        finalize_client(client);
    }

    #[test]
    #[cfg(any(feature = "with-s3-ci", feature = "with-containers"))]
    fn should_print_working_directory() {
        crate::mock::logger();
        let Ctx {
            mut client,
            container: _container,
        } = setup_client();
        assert!(client.pwd().is_ok());
        finalize_client(client);
    }

    #[test]
    #[cfg(any(feature = "with-s3-ci", feature = "with-containers"))]
    fn should_remove_dir_all() {
        crate::mock::logger();
        let Ctx {
            mut client,
            container: _container,
        } = setup_client();
        // Create dir
        let mut dir_path = client.pwd().ok().unwrap();
        dir_path.push(Path::new("test/"));
        assert!(
            client
                .create_dir(dir_path.as_path(), UnixPex::from(0o775))
                .is_ok()
        );
        // Create file
        let mut file_path = dir_path.clone();
        file_path.push(Path::new("a.txt"));
        let file_data = "test data\n";
        let reader = Cursor::new(file_data.as_bytes());
        let mut metadata = Metadata::default();
        metadata.size = file_data.len() as u64;
        assert!(
            client
                .create_file(file_path.as_path(), &metadata, Box::new(reader))
                .is_ok()
        );
        // Remove dir
        assert!(client.remove_dir_all(dir_path.as_path()).is_ok());
        finalize_client(client);
    }

    #[test]
    #[cfg(any(feature = "with-s3-ci", feature = "with-containers"))]
    fn should_remove_dir() {
        crate::mock::logger();
        let Ctx {
            mut client,
            container: _container,
        } = setup_client();
        // Create dir
        let mut dir_path = client.pwd().ok().unwrap();
        dir_path.push(Path::new("test/"));
        assert!(
            client
                .create_dir(dir_path.as_path(), UnixPex::from(0o775))
                .is_ok()
        );
        assert!(client.remove_dir(dir_path.as_path()).is_ok());
        finalize_client(client);
    }

    #[test]
    #[cfg(any(feature = "with-s3-ci", feature = "with-containers"))]
    fn should_not_remove_dir() {
        crate::mock::logger();
        let Ctx {
            mut client,
            container: _container,
        } = setup_client();
        // Remove dir
        assert!(client.remove_dir(Path::new("test/")).is_err());
        finalize_client(client);
    }

    #[test]
    #[cfg(any(feature = "with-s3-ci", feature = "with-containers"))]
    fn should_remove_file() {
        crate::mock::logger();
        let Ctx {
            mut client,
            container: _container,
        } = setup_client();
        // Create file
        let p = Path::new("a.txt");
        let file_data = "test data\n";
        let reader = Cursor::new(file_data.as_bytes());
        let mut metadata = Metadata::default();
        metadata.size = file_data.len() as u64;
        assert!(client.create_file(p, &metadata, Box::new(reader)).is_ok());
        assert!(client.remove_file(p).is_ok());
        // stat
        assert!(client.stat(p).is_err());
        finalize_client(client);
    }

    #[test]
    #[cfg(any(feature = "with-s3-ci", feature = "with-containers"))]
    fn should_not_setstat_file() {
        crate::mock::logger();
        let Ctx {
            mut client,
            container: _container,
        } = setup_client();
        // Create file
        let p = Path::new("a.sh");
        let file_data = "echo 5\n";
        let reader = Cursor::new(file_data.as_bytes());
        let mut metadata = Metadata::default();
        metadata.size = file_data.len() as u64;
        assert!(client.create_file(p, &metadata, Box::new(reader)).is_ok());
        assert!(
            client
                .setstat(
                    p,
                    Metadata {
                        accessed: Some(SystemTime::UNIX_EPOCH),
                        created: Some(SystemTime::UNIX_EPOCH),
                        gid: Some(1000),
                        file_type: remotefs::fs::FileType::File,
                        mode: Some(UnixPex::from(0o755)),
                        modified: Some(SystemTime::UNIX_EPOCH),
                        size: 7,
                        symlink: None,
                        uid: Some(1000),
                    }
                )
                .is_err()
        );
        finalize_client(client);
    }

    #[test]
    #[cfg(any(feature = "with-s3-ci", feature = "with-containers"))]
    fn should_stat_file() {
        crate::mock::logger();
        let Ctx {
            mut client,
            container: _container,
        } = setup_client();
        // Create file
        let p = Path::new("a.sh");
        let file_data = "echo 5\n";
        let reader = Cursor::new(file_data.as_bytes());
        let mut metadata = Metadata::default();
        metadata.size = file_data.len() as u64;
        assert!(client.create_file(p, &metadata, Box::new(reader)).is_ok());
        let entry = client.stat(p).ok().unwrap();
        assert_eq!(entry.name(), "a.sh");
        let mut expected_path = client.pwd().ok().unwrap();
        expected_path.push("a.sh");
        assert_eq!(entry.path(), expected_path.as_path());
        let meta = entry.metadata();
        assert_eq!(meta.size, 7);
        finalize_client(client);
    }

    #[test]
    #[cfg(any(feature = "with-s3-ci", feature = "with-containers"))]
    fn should_not_stat_file() {
        crate::mock::logger();
        let Ctx {
            mut client,
            container: _container,
        } = setup_client();
        // Create file
        let p = Path::new("a.sh");
        assert!(client.stat(p).is_err());
        finalize_client(client);
    }

    #[test]
    #[cfg(any(feature = "with-s3-ci", feature = "with-containers"))]
    fn should_make_symlink() {
        crate::mock::logger();
        let Ctx {
            mut client,
            container: _container,
        } = setup_client();
        // Create file
        let p = Path::new("a.sh");
        let file_data = "echo 5\n";
        let reader = Cursor::new(file_data.as_bytes());
        let mut metadata = Metadata::default();
        metadata.size = file_data.len() as u64;
        assert!(client.create_file(p, &metadata, Box::new(reader)).is_ok());
        let symlink = Path::new("b.sh");
        assert!(client.symlink(symlink, p).is_err());
        finalize_client(client);
    }

    #[test]
    #[cfg(any(feature = "with-s3-ci", feature = "with-containers"))]
    fn should_not_make_symlink() {
        crate::mock::logger();
        let Ctx {
            mut client,
            container: _container,
        } = setup_client();
        // Create file
        let p = Path::new("a.sh");
        let file_data = "echo 5\n";
        let reader = Cursor::new(file_data.as_bytes());
        let mut metadata = Metadata::default();
        metadata.size = file_data.len() as u64;
        assert!(client.create_file(p, &metadata, Box::new(reader)).is_ok());
        let symlink = Path::new("b.sh");
        let file_data = "echo 5\n";
        let reader = Cursor::new(file_data.as_bytes());
        assert!(
            client
                .create_file(symlink, &metadata, Box::new(reader))
                .is_ok()
        );
        assert!(client.symlink(symlink, p).is_err());
        assert!(client.remove_file(symlink).is_ok());
        assert!(client.symlink(symlink, Path::new("c.sh")).is_err());
        finalize_client(client);
    }

    #[test]
    fn should_return_errors_on_uninitialized_client() {
        let mut client =
            AwsS3Fs::new("aws-s3-test", &Arc::new(Runtime::new().unwrap())).region("eu-central-1");
        assert!(client.change_dir(Path::new("/tmp")).is_err());
        assert!(
            client
                .copy(Path::new("/nowhere"), PathBuf::from("/culonia").as_path())
                .is_err()
        );
        assert!(client.exec("echo 5").is_err());
        assert!(client.disconnect().is_err());
        assert!(client.symlink(Path::new("/a"), Path::new("/b")).is_err());
        assert!(client.list_dir(Path::new("/tmp")).is_err());
        assert!(
            client
                .create_dir(Path::new("/tmp"), UnixPex::from(0o755))
                .is_err()
        );
        assert!(client.pwd().is_err());
        assert!(client.remove_dir_all(Path::new("/nowhere")).is_err());
        assert!(
            client
                .mov(Path::new("/nowhere"), Path::new("/culonia"))
                .is_err()
        );
        assert!(client.stat(Path::new("/tmp")).is_err());
        assert!(
            client
                .setstat(Path::new("/tmp"), Metadata::default())
                .is_err()
        );
        assert!(client.open(Path::new("/tmp/pippo.txt")).is_err());
        assert!(
            client
                .create(Path::new("/tmp/pippo.txt"), &Metadata::default())
                .is_err()
        );
        assert!(
            client
                .append(Path::new("/tmp/pippo.txt"), &Metadata::default())
                .is_err()
        );
    }

    fn is_send<T: Send>(_send: T) {}

    fn is_sync<T: Sync>(_sync: T) {}

    #[test]
    fn test_should_be_sync() {
        let client = AwsS3Fs::new("bucket", &Arc::new(Runtime::new().unwrap()));

        is_sync(client);
    }

    #[test]
    fn test_should_be_send() {
        let client = AwsS3Fs::new("bucket", &Arc::new(Runtime::new().unwrap()));

        is_send(client);
    }

    // -- test utils

    #[allow(dead_code)]
    struct Ctx {
        client: AwsS3Fs,
        #[cfg(feature = "with-containers")]
        container: Minio,
        #[cfg(all(feature = "with-s3-ci", not(feature = "with-containers")))]
        container: (),
    }

    #[cfg(all(feature = "with-s3-ci", not(feature = "with-containers")))]
    fn setup_client() -> Ctx {
        // Get transfer
        let bucket = env!("AWS_S3_BUCKET");
        let mut client = AwsS3Fs::new(bucket, &Arc::new(Runtime::new().unwrap()));
        assert!(client.connect().is_ok());
        // Create wrkdir
        let tempdir = PathBuf::from(generate_tempdir());
        assert!(
            client
                .create_dir(tempdir.as_path(), UnixPex::from(0o775))
                .is_ok()
        );
        // Change directory
        let err = client.change_dir(tempdir.as_path());
        if err.is_err() {
            println!("Error: {:?}", err);
        }
        assert!(client.change_dir(tempdir.as_path()).is_ok());
        Ctx {
            client,
            container: (),
        }
    }

    #[cfg(feature = "with-containers")]
    fn setup_client() -> Ctx {
        let minio = Minio::start();
        let port = minio.port();

        // Get transfer
        let runtime = Arc::new(Runtime::new().expect("Could not create runtime"));
        let mut client = AwsS3Fs::new("github-ci", &runtime)
            .endpoint(format!("http://localhost:{port}"))
            .access_key("minioadmin")
            .secret_access_key("minioadmin")
            .new_path_style(true);

        // connect
        assert!(client.connect().is_ok());

        // Create bucket manually
        let fut = client
            .client()
            .unwrap()
            .create_bucket()
            .bucket("github-ci")
            .send();
        let res = runtime.block_on(fut);

        assert!(res.is_ok());

        // Create wrkdir
        let tempdir = PathBuf::from(generate_tempdir());
        assert!(
            client
                .create_dir(tempdir.as_path(), UnixPex::from(0o775))
                .is_ok()
        );
        // Change directory
        assert!(client.change_dir(tempdir.as_path()).is_ok());
        Ctx {
            client,
            container: minio,
        }
    }

    #[cfg(any(feature = "with-s3-ci", feature = "with-containers"))]
    fn finalize_client(mut client: AwsS3Fs) {
        // Get working directory
        let wrkdir = client.pwd().ok().unwrap();
        // Remove directory
        assert!(client.remove_dir_all(wrkdir.as_path()).is_ok());
        assert!(client.disconnect().is_ok());
    }

    #[cfg(any(feature = "with-s3-ci", feature = "with-containers"))]
    fn generate_tempdir() -> String {
        use rand::distr::Alphanumeric;
        use rand::{Rng, rng};
        let mut rng = rng();
        let name: String = std::iter::repeat(())
            .map(|()| rng.sample(Alphanumeric))
            .map(char::from)
            .take(8)
            .collect();
        format!("/github-ci/temp_{}/", name)
    }
}
